-- search.e

include rfont.e
include board.e
include eval.e

-- /* sort_pv() is called when the search function is following
--    the PV (Principal Variation). It looks through the current
--    ply's move list to see if the PV move is there. If so,
--    it adds 10,000,000 to the move's score so it's played first
--    by the search function. If not, follow_pv remains FALSE and
--    search() stops calling sort_pv(). */

procedure sort_pv()

    follow_pv=FALSE
    for i=gen_begin[ply] to gen_end[ply]-1  do 
        if equal(gen_dat[i][MOVE_],pv[ply][i]) then
            follow_pv=TRUE
            gen_dat[i][SCORE] += 10000000 
            return
        end if 
    end for

end procedure


-- /* sort() searches the current ply's move list from 'from'
--    to the end to find the move with the highest score. Then it
--    swaps that move and the 'from' move so the move with the
--    highest score gets searched next, and hopefully produces
--    a cutoff. */

procedure sort(integer from)
    sequence  x
    integer bs   --/* best score */
    integer bi   --/* best i */

    bs=-1 
    bi = 1

    for i=from  to gen_end[ply]-1 do
        if gen_dat[i][SCORE]>bs then
            bs=gen_dat[i][SCORE]
            bi=i 
        end if
    end for

    x=gen_dat[from] 
    gen_dat[from]=gen_dat[bi] 
    gen_dat[bi]=x

end procedure


-- /* quiesce() is a recursive minimax search function with
--    alpha-beta cutoffs. In other words, negamax. It basically
--    only searches capture sequences and allows the evaluation
--    Function to cut the search off (and set alpha). The idea
--    is to find a position where there isn't a lot going on
--    so the static evaluation function will work. */

function quiesce(integer alpha,integer beta)

    integer x 
    integer c   --/* in check */
    integer f   --/* legal move found */

    nodes += 1 

    pv_length[ply]=ply 

    c=in_check(side) 

    --/* if we're in check, try all moves to get out. (not
    --   necessary, just something I (Tom) decided to do) */

    if c then
        gen() 
    else --/* otherwise, use the evaluation function. */
        x=eval() 
        if x>=beta then
            return beta
        end if
        if x>alpha then
            alpha=x 
        end if
        gen_caps() 
    end if

    if follow_pv then  --/* are we following the PV? */
        sort_pv() 
    end if

    f=FALSE 

    --/* loop through the moves */

    for i=gen_begin[ply] to gen_end[ply]-1 do
        sort(i) 
        if makemove(gen_dat[i][MOVE_]) then
            f=TRUE   --/* we found a legal move! */
            x=-quiesce(-beta,-alpha) 
            takeback() 
            if x>alpha then
                if x>=beta then
                    return beta
                end if
                alpha=x 
                --/* update the PV */
                pv[ply][ply]=gen_dat[i][MOVE_]
                for j=ply+1 to pv_length[ply+1]-1 do
                    pv[ply][j]=pv[ply+1][j] 
                end for
                pv_length[ply]=pv_length[ply+1] 
            end if
        end if 
    end for

    --/* if we're in check and there aren't any legal moves,
    --   well, we lost */

    if not f and c then
        return -10000+ply-1
    end if

    return alpha 

end function 


-- /* search() does just that, in negamax fashion */
global
function search(integer alpha,integer beta,integer depth)

    integer x 
    integer c,f 

    --/* we're as deep as we want to be  call quiesce() to get
    --   a reasonable score and return it. */

    if not depth then
        return quiesce(alpha,beta)
    end if

    nodes += 1 

    pv_length[ply]=ply

    --/* are we in check? if so, we want to search deeper */

    c=in_check(side) 

    if c then
        depth += 1
    end if 

    gen() 

    if follow_pv then  --/* are we following the PV? */
        sort_pv() 
    end if

    f=FALSE 

    --/* loop through the moves */

    for i=gen_begin[ply] to gen_end[ply]-1 do
        sort(i) 
        if makemove(gen_dat[i][MOVE_]) then 
            f=TRUE 
            x=-search(-beta,-alpha,depth-1) 
            takeback() 
            if x>alpha then
                --/* this move caused a cutoff, so increase the history
                --   value so it gets ordered high next time we can
                --   search it */
                history[gen_dat[i][MOVE_][FROM]][gen_dat[i][MOVE_][TO]]+=depth 
                if x>=beta then
                    return beta
                end if
                alpha=x 
                --/* update the PV */
                pv[ply][ply]=gen_dat[i][MOVE_] 
                for j=ply+1 to pv_length[ply+1] do
                    pv[ply][j]=pv[ply+1][j] 
                end for
                pv_length[ply]=pv_length[ply+1] 
            end if
        end if
    end for

    --/* no legal moves? then we're in checkmate or stalemate */

    if not f then
        if c then
            return -10000+ply-1
        else
            return 0
        end if
    end if

    -- /* fifty move draw rule */

    if fifty>=100 then
        return 0 
    end if

    return alpha

end function


--/* think() calls search() iteratively and prints the results
--   after every iteration. */

global
procedure think()
    integer x 
    ply=1
    nodes=0
    polygon(CYAN,1,{{50,370},{639,370},{639,479},{50,479}})
    set_foreground(0)
    setx(50) sety(370)
    write("ply      nodes  score  pv")
    for i=1 to DEPTHMAX do 
        follow_pv=TRUE 
        x=search(-10000,10000,i) 
        setx(50) sety(370+i*15)
        write(sprintf("%3d  %9d  %5d ",{i,nodes,x}))
        for j=1 to pv_length[1]-1 do
            write(sprintf(" %s%s%s%s",
                  {FILE_[pv[1][j][FROM]]+'a'- 1,
                  9-RANK_[pv[1][j][FROM]]+'0',
                  FILE_[pv[1][j][TO]]+'a' - 1,
                  9-RANK_[pv[1][j][TO]]+'0'})) 
        end for                            
        if x >= 9999 or x <= -10000 then
           gameover = 1
           exit 
        end if
    end for
end procedure

global
procedure hint()
    integer x 
    ply=1
    nodes=0
    polygon(CYAN,1,{{50,370},{639,370},{639,479},{50,479}})
    set_foreground(0)
    setx(50) sety(370)
    write("ply      nodes  score  hints")
    for i=1 to DEPTHMAX do 
        follow_pv=TRUE 
        x=search(-10000,10000,i) 
        setx(50) sety(370+i*15)
        write(sprintf("%3d  %9d  %5d ",{i,nodes,x}))
        for j=1 to pv_length[1]-1 do
            write(sprintf(" %s%s%s%s",
                  {FILE_[pv[1][j][FROM]]+'a'- 1,
                  9-RANK_[pv[1][j][FROM]]+'0',
                  FILE_[pv[1][j][TO]]+'a' - 1,
                  9-RANK_[pv[1][j][TO]]+'0'})) 
        end for                            
    end for
end procedure

